#!/bin/sh


WORK=ASPEN.K2-1.5p8


export ASPEN_TUNING_LEVEL=DEFAULT
export ASPEN_REGUSE_CALC=0
export ASPEN_CHECK_ERROR=0

export ASPEN_GPU_ID=1

if [ x"$ASPEN_GPU_ID" = x ]; then
	export ASPEN_GPU_ID=0
fi

BUILD=`date '+%Y%m%d.%H%M%S'`

LD_PATH_OLD=$LD_LIBRARY_PATH

export CUDA_PATH=/usr/local/cuda
if [ x"$LD_PATH_OLD" = x ]; then
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64
else
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_PATH_OLD
fi

if [ ! -e $WORK ]; then
	if [ ! -e $WORK.tgz ]; then
		echo 'ERROR : '$WORK'.tgz does not exist.'
		exit 1
	else
		tar zxvf $WORK.tgz
	fi
fi


cd $WORK
make clobber

CUDA=8000
VERSION=`echo $CUDA | awk '{ v=$1; s=int(v/100); m=int(v-s*100); s/=10; m/=10; print s"."m; }'`


export CUDA_PATH=/usr/local/cuda-$VERSION
if [ x"$LD_PATH_OLD" = x ]; then
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64
else
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_PATH_OLD
fi
make
export CUDAx=`awk '/CURRENT_GPU/{ print $3; }' tuning/CURRENT_GPU`
export GPU=`awk '/DEVICE=/{ print $2; }' tuning/DEV_INFO`

cd ..


export ASPEN_TUNING_LEVEL=DEFAULT

if [ $CUDAx -lt 600 ]; then
	CUDA_list='8000 7050 7000'
else
	CUDA_list='8000'
fi

for CUDA in \
	$CUDA_list
do

	VERSION=`echo $CUDA | awk '{ v=$1; s=int(v/100); m=int(v-s*100); s/=10; m/=10; print s"."m; }'`

	export CUDA_PATH=/usr/local/cuda-$VERSION
if [ x"$LD_PATH_OLD" = x ]; then
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64
else
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_PATH_OLD
fi
	cd $WORK
	make clobber
	make
	export GPU=`awk '/DEVICE=/{ print $2; }' tuning/DEV_INFO`
	make tuned-new-symv |& tee log-newkernel-$GPU-$CUDA
	cd tuning
	if [ ! -e tuning-old/$GPU ]; then
		mkdir tuning-old/$GPU
	fi
	if [ ! -e tuning-old/$GPU/$CUDA ]; then
		mkdir tuning-old/$GPU/$CUDA
	fi
	cp *.c *.h tuning-old/$GPU/$CUDA/
	cd ../..

export ASPEN_TUNING_LEVEL=ROUGH

done


export CUDA_PATH=/usr/local/cuda
if [ x"$LD_PATH_OLD" = x ]; then
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64
else
	export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_PATH_OLD
fi

cd $WORK
export GPU=`awk '/DEVICE=/{ print $2; }' tuning/DEV_INFO`
cd ..


tar -C $WORK/tuning/tuning-old -zcvf $WORK-$GPU-$BUILD.tgz $GPU
tar zcvf $WORK-$BUILD.tgz $WORK
mv $WORK $WORK-$BUILD
tar zxvf $WORK-$BUILD.tgz
cd $WORK
make clobber
\rm log-*


make
cd bench
make
export XGPU=`echo $GPU | awk '/DEVICE/{ gsub(/^[A-Za-z0-9]*-/,"",$2); gsub(/-/,""); print $2; }'`
/bin/sh ./exe.sh
if [ ! -e logs ]; then
	mkdir logs
fi
mv log-*$XGPU-* logs/
cd ..
make clobber


cd ..
tar zcvf $WORK.tgz $WORK


export LD_LIBRARY_PATH=$LD_PATH_OLD


exit

