#ifndef ASPEN_ATOMIC_H_INCLUDED

#if GPU_ARCH >= Fermi

//
// atomic Operations are supported on Fermi or later GPU generations
// 
#if GPU_ARCH < Pascal
__forceinline__
__device__ double atomicAdd ( double* address, double val )
{
  unsigned long long int * addr = (unsigned long long int*)address;
  unsigned long long int old    = * addr, assumed;

  do {
    assumed = old;
    const unsigned long long int update = 
      __double_as_longlong(val + __longlong_as_double(old));
    old = atomicCAS(addr, assumed, update );
  } while (assumed != old);

  return __longlong_as_double(old);
}
#endif

#if GPU_ARCH < Pascal
__forceinline__
__device__ double atomicExch ( double * address, double val )
{
#if GPU_ARCH < Kepler2
  unsigned long long int * addr = (unsigned long long int*)address;
  unsigned long long int old    = * addr, assumed;
  const unsigned long long int update = __double_as_longlong(val);

  do {
    assumed = old;
    old = atomicCAS(addr, assumed, update );
  } while (assumed != old);

  return __longlong_as_double(old);
#else
  unsigned long long  * addr = (unsigned long long  *)address;
  return __longlong_as_double(atomicExch(addr, __double_as_longlong(val)));
#endif
}
#endif

#if GPU_ARCH < Kepler2
__forceinline__
__device__ unsigned long long atomicOr__ ( unsigned long long * address, unsigned long long val )
{
  unsigned long long * addr = (unsigned long long *)address;
  unsigned long long old    = * addr, assumed, update;
  do {
    assumed = old;
    update = old | val;
    old = atomicCAS( addr, assumed, update );
  } while (assumed != old);
  return old;
}
#endif

__forceinline__
__device__ double atomicLoad ( double * address )
{
  unsigned long long * addr = (unsigned long long *)address;
  unsigned long long   mask = 0;
#if GPU_ARCH < Kepler2
  return __longlong_as_double( atomicOr__( addr, mask ) );
#else
  return __longlong_as_double( atomicOr( addr, mask ) );
#endif
}

__forceinline__
__device__ float atomicLoad ( float * address )
{
  unsigned int * addr = (unsigned int *)address;
  unsigned int   mask = 0;
  unsigned int   t = atomicOr( addr, mask );
  float          s;
#if CUDA_VERSION<7050
  asm volatile ( "mov.f32\t%0, %1;" : "=f"(s) : "r"(t) );
#else
  s = __uint_as_float( t );
#endif
  return s;
}

#endif


#define ASPEN_ATOMIC_H_INCLUDED		1
#endif

