// additional including standard header files
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <ctype.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>


// additional including CUDA header files
#include <cuda_runtime.h>
#include <cuda.h>
#if CUDA_VERSION<5000
#   include <cutil.h>
#endif
#include <cublas.h>


#ifndef ASPEN_DEVEL_H_INCLUDED

// definitions for generic macros
#define _MACRO_WRAP_(...)                       \
  do { { __VA_ARGS__; } break; } while( 0 )
#define _MACRO_BLOCK_		_MACRO_WRAP_

#define __4PARAMS__(N,x,y,z,w,...)	N##_##x##_##y##_##z##_##w
#define __3PARAMS__(N,x,y,z,...)	N##_##x##_##y##_##z
#define __2PARAMS__(N,x,y,...)		N##_##x##_##y
#define	__1PARAMS__(N,x,...)		N##_##x

#define _NAME_4PARAMS_(N,x,y,z,w,...)	__4PARAMS__(N,x,y,z,w)
#define _NAME_3PARAMS_(N,x,y,z,...)	__3PARAMS__(N,x,y,z)
#define _NAME_2PARAMS_(N,x,y,...)	__2PARAMS__(N,x,y)
#define	_NAME_1PARAMS_(N,x,...)		__1PARAMS__(N,x)

// definitions for the prefix of API
#define	API_PREFIX	ASPEN
#define	API(x,...)	_NAME_1PARAMS_(API_PREFIX,x)



// CUDA architecture CONST
#define	Tesla		(130)
#define	Fermi		(200)
#define	Kepler		(300)
#define	Kepler2		(350)
#define	Maxwell		(500)
#define	Maxwell2	(520)
#define	Maxwell3	(530)
#define	Pascal		(600)
#define	Pascal1		(610)
#define	Pascal2		(620)


#define	ASPEN_GPU_Names(x,...)                                          \
  (                                                                     \
  ( ( x == Tesla )        ? "Tesla"                                     \
  : ( ( x == Fermi )    ? "Fermi"                                       \
  : ( ( x == Kepler )   ? "Kepler"                                      \
      : ( ( x == Kepler2 )  ? "Kepler2"                                 \
          : ( ( x == Maxwell )  ? "Maxwell"                             \
              : ( ( x == Maxwell2 ) ? "Maxwell2"                        \
                  : ( ( x == Maxwell3 ) ? "Maxwell3"                    \
                      : ( ( x == Pascal )   ? "Pascal"                  \
                          : ( ( x == Pascal1 )   ? "Pascal1"            \
                              : ( ( x == Pascal2 )   ? "Pascal2"        \
                                  : ( "unknown" ) ) ) ) ) ) ) )         \
      )


//
// The macro __CUDA_ARCH__ is only available on the compilation for devices
// To use __CUDA_ARCH__ sometimes confuses us, so we separately compile
// the source code xxxx_X-template.cu by nvcc option such as
// --generate-code arch=compute_XX,code=compute_XX, and GPU_ARCH can be
// used for the distingishment of the GPU kernel.
//

#include "aspen_gpu_conf.h"



// Error check and safe invocation of the kernel function
#if 0
#define	ASPEN_KERNEL_INVOKE( kernel_name, ...)				\
  _MACRO_BLOCK_ (							\
                 kernel_name  __VA_ARGS__ ;				\
                 cudaError_t err = cudaGetLastError();			\
                 if ( err != cudaSuccess ) {				\
                   printf ( "ERROR in [%s] : %s \n",                    \
                            #kernel_name, cudaGetErrorString(err) );    \
                   exit ( 1 );                                          \
                   return;						\
                 }							\
                                                                        )
#else
#define	ASPEN_KERNEL_INVOKE( kernel_name, ...)				\
  _MACRO_BLOCK_ (							\
                 kernel_name  __VA_ARGS__ ;				\
                                                                        )
#endif
#define __LOOP_DIVIDER__(...)                   \
  asm( "// <- nop ->;" )


#include "aspen_texture.h"
#include "aspen_complex.h"
#include "aspen_const.h"
#if 0
#include "aspen_atomic.h"
#endif
#include "aspen_ld_st.h"
#include "aspen_istypes.h"



#include "aspen_devel_func.h"

/* Utility functions for C APIs */
extern "C"  int     API(get_numbers_MP)      ( void );
extern "C"  void    API(get_device_Name)     ( char * name );
extern "C"  int     API(get_device_Major)    ( void );
extern "C"  int     API(get_device_Minor)    ( void );
extern "C"  int     API(get_device_CC)       ( void );
extern "C"  int     API(get_driver_Version)  ( void );
extern "C"  int     API(get_runtime_Version) ( void );
extern "C"  size_t  API(get_device_Memory)   ( void );
extern "C"  size_t  API(get_device_WorkSize) ( void );
extern "C"  void *  API(get_device_WorkArea) ( void );

#include "aspen.h"

#define ASPEN_DEVEL_H_INCLUDED		1
#endif

