#include "aspen_devel.h"

int print_dev_info ( const char *device_name,
                     const int MPs, const int CG, const size_t MMEM,
		     const size_t MAXDIM[] );

int
main( int argc, char *argv[] )
{
  int	id = 0;
  if ( argc > 1 ) {
    sscanf( argv[1], "%d", &id );
  } else {
    char *id_str = getenv("ASPEN_GPU_ID");
    if ( id_str ) {
      sscanf( id_str, "%d", &id );
    }
  }

  cudaError_t err = cudaSetDevice( id );
  if ( err != cudaSuccess ) {
    size_t	MAXDIM[4] = { 0, 0, 0, 0};
    perror("CUDA device open FAILED");
    print_dev_info ( "unknown", (int)0, (int)0, (size_t)0, &MAXDIM[0] );
    return EXIT_FAILURE;
  }

  ASPEN_init   ( id );

  char	device_name[1024]; ASPEN_get_device_Name( device_name );
  int	MPs     = ASPEN_get_numbers_MP( );
  int	CG      = ASPEN_get_device_CC( );
  size_t	MMEM    = ASPEN_get_device_Memory( );
  size_t	MAXDIM[4];
  MMEM -= ASPEN_get_device_WorkSize( );

  {
    char *s = device_name;
    while ( (s = strchr( s, ' ' )) != NULL ) { *s = '-'; }
  }

  MAXDIM[0] = MMEM / (sizeof(double));
  MAXDIM[0] = (size_t)(sqrt((double)MAXDIM[0])*0.95);

  MAXDIM[1] = MMEM / (sizeof(float));
  MAXDIM[1] = (size_t)(sqrt((double)MAXDIM[1])*0.95);

  MAXDIM[2] = MMEM / (sizeof(double)*2);
  MAXDIM[2] = (size_t)(sqrt((double)MAXDIM[2])*0.95);

  MAXDIM[3] = MMEM / (sizeof(double)*4);
  MAXDIM[3] = (size_t)(sqrt((double)MAXDIM[3])*0.95);

  print_dev_info ( device_name, MPs, CG, MMEM, MAXDIM );

  printf( "// cuda version\n" );
  printf( "CUDA= %d\n",  ASPEN_get_runtime_Version() );

  return EXIT_SUCCESS;
}

int
print_dev_info ( const char *device_name,
                 const int MPs, const int CG, const size_t MMEM,
		 const size_t MAXDIM[] )
{
  printf( "// device name\n" );
  printf( "DEVICE= %s\n"   , device_name );
  printf( "// the number of multi-processors\n" );
  printf( "MP= %d\n"       , MPs         );
  printf( "// compute-compatibility generation\n" );
  printf( "CG= %d\n"       , CG          );
  printf( "// capacity of the global memory\n" );
  printf( "MAXmem= %ld\n"  , MMEM        );
  printf( "// for double or cuFloatComplex\n" );
  printf( "MAXDIM= %ld\n"  , MAXDIM[0]   );
  printf( "// for float\n" );
  printf( "MAXDIM2= %ld\n" , MAXDIM[1]   );
  printf( "// for cuFloatComplex or DD\n" );
  printf( "MAXDIM3= %ld\n" , MAXDIM[2]   );
  printf( "// for DD-Complex\n" );
  printf( "MAXDIM4= %ld\n" , MAXDIM[3]   );
}

