#if GPU_ARCH>=Kepler

template < class TYPE >
__forceinline__ __device__ TYPE
__SHFL_xor ( TYPE const x, int const laneMask, int const width )
{
  return  __shfl_xor( x, laneMask, width );
}

template < >
__forceinline__ __device__ double
__SHFL_xor < double > ( double const x, int const laneMask, int const width )
{
  int tmp_hi, tmp_lo;
#if 1
  asm volatile ( "mov.b64  {%1, %0}, %2;\n\t"
                 :"=r"(tmp_hi), "=r"(tmp_lo)
                 :"d"(x) );
#else
  tmp_hi = __double2hiint( x );
  tmp_lo = __double2loint( x );
#endif
  tmp_hi = __shfl_xor( tmp_hi, laneMask, width );
  tmp_lo = __shfl_xor( tmp_lo, laneMask, width );
  return  __hiloint2double( tmp_hi, tmp_lo );
}

template < >
__forceinline__ __device__ cuFloatComplex
__SHFL_xor < cuFloatComplex > ( cuFloatComplex const v, int const laneMask, int const width )
{
  float tmp_x, tmp_y;
  tmp_x = __shfl_xor( v.x, laneMask, width );
  tmp_y = __shfl_xor( v.y, laneMask, width );
  return  make_cuFloatComplex ( tmp_x, tmp_y );
}

template < >
__forceinline__ __device__ cuDoubleComplex
__SHFL_xor < cuDoubleComplex > ( cuDoubleComplex const v, int const laneMask, int const width )
{
  int tmp_x_hi, tmp_x_lo;
  int tmp_y_hi, tmp_y_lo;
#if 1
  asm volatile ( "mov.b64  {%1, %0}, %2;\n\t"
                 :"=r"(tmp_x_hi), "=r"(tmp_x_lo)
                 :"d"(v.x) );
  asm volatile ( "mov.b64  {%1, %0}, %2;\n\t"
                 :"=r"(tmp_y_hi), "=r"(tmp_y_lo)
                 :"d"(v.y) );
#else
  tmp_x_hi = __double2hiint( v.x );
  tmp_x_lo = __double2loint( v.x );
  tmp_y_hi = __double2hiint( v.y );
  tmp_y_lo = __double2loint( v.y );
#endif
  tmp_x_hi = __shfl_xor( tmp_x_hi, laneMask, width );
  tmp_x_lo = __shfl_xor( tmp_x_lo, laneMask, width );
  tmp_y_hi = __shfl_xor( tmp_y_hi, laneMask, width );
  tmp_y_lo = __shfl_xor( tmp_y_lo, laneMask, width );
  return  make_cuDoubleComplex ( __hiloint2double( tmp_x_hi, tmp_x_lo ), __hiloint2double( tmp_y_hi, tmp_y_lo ) );
}


// In return, if flag is true, x and y are swapped
//  int flag : flag whether swap the arguments
//  TYPE x,y : regsiters
#define	Swap_Elements(x,y,flag, ... )                                   \
  _MACRO_BLOCK_(                                                        \
		const bool FLAG = (bool) flag;                          \
                TYPE  xx = x;                                           \
                x = FLAG ? y : xx;                                      \
                y = FLAG ? xx : y;                                      \
                                                                        )
// In return, warp-oriented sumup are stored on x0 of the first warp
// Hidden / un-passed/ implicit parameters
//  TYPE x0  : variable to be sumed up
//  TYPE w[] : work area on shared_memory
//  int  c = local_id / 32; : the warp number
#define	_sumup_inter_Warp( ... )                                        \
  _MACRO_BLOCK_(                                                        \
                switch ( BLOCK_SIZE ) {                                 \
                case 32:                                                \
                  break;                                                \
                case 64:                                                \
                  if ( c ) { w[0] = x0; }                               \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32]; }                           \
                  break;                                                \
                case 96:                                                \
                  if ( c ) { w[0] = x0; }                               \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32] + w[64]; }                   \
                  break;                                                \
                case 128:                                               \
                  if ( c >= 2 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c < 2 ) { w[0] = (x0 += w[64]); }                \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32]; }                           \
                  break;                                                \
                case 160:                                               \
                  if ( c >= 3 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c*(c-3) < 0 ) { w[0] = x0 + w[64]; }             \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32] + w[64]; }                   \
                  break;                                                \
                case 192:                                               \
                  if ( c > 2 ) { w[0] = x0; }                           \
                  __syncthreads ( );                                    \
                  if ( c <= 2 ) { w[0] = (x0 += w[96]); }               \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32] + w[64]; }                   \
                  break;                                                \
                case 224:                                               \
                  if ( c >= 4 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c*(c-4) < 0 ) { w[0] = (x0 += w[96]); }          \
                  __syncthreads ( );                                    \
                  if ( c < 2 ) { w[0] = (x0 += w[64]); }                \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32]; }                           \
                  break;                                                \
                case 256:                                               \
                  if ( c >= 4 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c < 4 ) { w[0] = (x0 += w[128]); }               \
                  __syncthreads ( );                                    \
                  if ( c < 2 ) { w[0] = (x0 += w[64]); }                \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32]; }                           \
                  break;                                                \
                case 288:                                               \
                  if ( c >= 3 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c < 3 ) { w[0] = (x0 += w[96] + w[192]); }       \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32] + w[64]; }                   \
                  break;                                                \
                case 320:                                               \
                  if ( c >= 4 ) { w[0] = x0; }                          \
                  __syncthreads ( );                                    \
                  if ( c*(c-4) < 0 ) { w[0] = (x0 += w[128] + w[256]); } \
                  __syncthreads ( );                                    \
                  if ( c < 2 ) { w[0] = (x0 += w[64]); }                \
                  __syncthreads ( );                                    \
                  if ( ! c ) { x0 += w[32]; }                           \
                  break;                                                \
                default:                                                \
                  if ( c ) { w[0] = x0; }                               \
                  __syncthreads ( );                                    \
                  if ( c == 0 ) {                                       \
                    __volatile__                                        \
                      TYPE  * W = w + 32;                               \
                    for ( int i=32; i<BLOCK_SIZE; i+=32 ) {             \
                      x0 += (* W); W += 32;                             \
                    }                                                   \
                  }                                                     \
                  break;                                                \
                }                                                       \
                                                                        )

// In return, x@mine += x@(mine^(1<<lane))
//  TYPE x   : regsiters to be summed up
//  int lane : lane , a distance of the registers sumed up
#define	ADD_Default( a, lane, ... )                                     \
  _MACRO_BLOCK_(                                                        \
                a += __SHFL_xor < TYPE > ( a,  lane, 32 );              \
                                                                        )

#define	ADD_WithSwap1( a, b, lane, ... )                                \
  _MACRO_BLOCK_(                                                        \
                const bool flag = (threadIdx.x & lane);                 \
                Swap_Elements( a, b, flag );                            \
                                                                        )
#define	ADD_WithSwap2( a, b, lane, ... )                                \
  _MACRO_BLOCK_(                                                        \
                a += __SHFL_xor < TYPE > ( b,  lane, 32 );              \
                                                                        )
// In return, x@{mine,(mine^(1<<lane))} += y@{(mine^(1<<lane)),mine}
//  TYPE x,y : regsiters to be summed up
//  int lane : lane , a distance of butterfly swaps
#define	ADD_WithSwap( a, b, lane, ... )                                 \
  _MACRO_BLOCK_(                                                        \
                ADD_WithSwap1( a, b, lane );                            \
                ADD_WithSwap2( a, b, lane );                            \
                                                                        )

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce_in_block ( TYPE x0, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 1  );
    ADD_Default( x0, 2  );
    ADD_Default( x0, 4  );
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x == 0 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce2_in_block ( TYPE x0, TYPE x1, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 2  );
    ADD_Default( x0, 4  );
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 2 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__ __device__ void
reduce3_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap( x0, x1, 1 );
    ADD_Default ( x2,  1 );
    ADD_WithSwap( x0, x2, 2 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 4  );
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 3 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__ __device__ void
reduce4_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_WithSwap ( x0, x2, 2 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 4  );
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 4 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__ __device__ void
reduce5_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_Default  ( x4,  1 );
    ADD_WithSwap ( x0, x2, 2 );
    ADD_Default  ( x4,  2 );
    ADD_WithSwap ( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 5 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce6_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap1( x4, x5, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_WithSwap2( x4, x5, 1 );
    ADD_WithSwap ( x0, x2, 2 );
    ADD_Default  ( x4, 2 );
    ADD_WithSwap ( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 6 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce7_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap1( x4, x5, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_WithSwap2( x4, x5, 1 );
    ADD_Default  ( x6, 1 );
    ADD_WithSwap1( x0, x2, 2 );
    ADD_WithSwap1( x4, x6, 2 );
    ADD_WithSwap2( x0, x2, 2 );
    ADD_WithSwap2( x4, x6, 2 );
    ADD_WithSwap ( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 7 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce8_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE x7, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap1( x4, x5, 1 );
    ADD_WithSwap1( x6, x7, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_WithSwap2( x4, x5, 1 );
    ADD_WithSwap2( x6, x7, 1 );
    ADD_WithSwap1( x0, x2, 2 );
    ADD_WithSwap1( x4, x6, 2 );
    ADD_WithSwap2( x0, x2, 2 );
    ADD_WithSwap2( x4, x6, 2 );
    ADD_WithSwap ( x0, x4, 4 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 8  );
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 8 ) z[0] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__INLINE__
__device__ void
reduce9_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE x7, TYPE x8, TYPE * w_, TYPE * z )
{

  const long local_id = threadIdx.x;
  const int  c = threadIdx.x / 32;
  __volatile__ TYPE  * w = (TYPE *)w_;
  w += local_id; z += local_id;

  if ( BLOCK_SIZE > 32 ) { __syncthreads ( ); }

  {
    ADD_WithSwap1( x0, x1, 1 );
    ADD_WithSwap1( x2, x3, 1 );
    ADD_WithSwap1( x4, x5, 1 );
    ADD_WithSwap1( x6, x7, 1 );
    ADD_WithSwap2( x0, x1, 1 );
    ADD_WithSwap2( x2, x3, 1 );
    ADD_WithSwap2( x4, x5, 1 );
    ADD_WithSwap2( x6, x7, 1 );
    ADD_Default  ( x8,  1 );
    ADD_WithSwap1( x0, x2, 2 );
    ADD_WithSwap1( x4, x6, 2 );
    ADD_WithSwap2( x0, x2, 2 );
    ADD_WithSwap2( x4, x6, 2 );
    ADD_Default  ( x8,  2 );
    ADD_WithSwap1( x0, x4, 4 );
    ADD_WithSwap2( x0, x4, 4 );
    ADD_Default  ( x8,  4 );
    ADD_WithSwap ( x0, x8, 8 );
  }

  if ( BLOCK_SIZE > 32 ) { _sumup_inter_Warp( ); }

  if ( c == 0 ) {
    ADD_Default( x0, 16 );
  }

  if ( threadIdx.x < 9 ) z[0] = x0;
}

#endif

